package org.bazaar_vcs.qbzr_eclipse;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.bazaar_vcs.qbzr_eclipse.util.BazaarUtilities;
import org.bazaar_vcs.qbzr_eclipse.util.FileUtilities;
import org.bazaar_vcs.qbzr_eclipse.util.ProcessException;
import org.bazaar_vcs.qbzr_eclipse.util.ProcessUtilities;
import org.eclipse.core.resources.IResourceChangeEvent;
import org.eclipse.core.resources.IResourceChangeListener;
import org.eclipse.core.resources.IResourceDelta;
import org.eclipse.core.resources.IResourceDeltaVisitor;
import org.eclipse.core.resources.IWorkspaceRoot;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;

/**
 * Listener that listens for changes to Eclipse resources and then invokes the
 * "bzr move" commands (for example, when refactoring in Eclipse).
 * 
 * @author nick
 * 
 */
public class ResourceChangeListener implements IResourceChangeListener
{
    public static final ResourceChangeListener INSTANCE = new ResourceChangeListener();

    private ResourceChangeListener()
    {
        mThread = new Thread()
        {
            @Override
            public void run()
            {
                processFileChangsQueue();
            }
        };

        mThread.start();
    }

    public void stop()
    {
        synchronized (mFileChangsQueue)
        {
            mShouldStop = true;
            mFileChangsQueue.notifyAll();
        }

        try
        {
            mThread.join();
        }
        catch (InterruptedException e)
        {
        }
    }

    /** Causes the calling thread to wait until the queue has been emptied . */
    public void waitForFilesToBeMovedAndDeleted()
    {
        synchronized (mFileChangsQueue)
        {
            if (mFileChangsQueue.isEmpty())
            {
                return;
            }
        }

        QBzrEclipsePlugin.getInstance().logInfo(
                "Waiting for Bazaar to add, delete, and move files");

        synchronized (mQueueEmptiedSignal)
        {
            try
            {
                mQueueEmptiedSignal.wait();
            }
            catch (InterruptedException e)
            {

            }
        }

        QBzrEclipsePlugin.getInstance().logInfo(
                "Bazaar finished adding, deleting, and moving files");
    }

    public void resourceChanged(IResourceChangeEvent event)
    {
        if (event.getType() == IResourceChangeEvent.POST_CHANGE)
        {
            IResourceDelta delta = event.getDelta();
            final FileChanges fileChanges = new FileChanges();

            try
            {
                delta.accept(new IResourceDeltaVisitor()
                {
                    public boolean visit(IResourceDelta delta)
                            throws CoreException
                    {
                        return processDelta(delta, fileChanges);
                    }
                });

                addFileChanges(fileChanges);
            }
            catch (CoreException e)
            {
                QBzrEclipsePlugin.getInstance().logException(
                        "Error processing resource change event", e);
            }
        }
    }

    private boolean processDelta(IResourceDelta delta, FileChanges fileChanges)
    {
        if (delta.getKind() == IResourceDelta.ADDED)
        {
            File addedFile = delta.getResource().getLocation().toFile();

            if ((delta.getFlags() & IResourceDelta.MOVED_FROM) != 0)
            {
                IWorkspaceRoot workspaceRoot = ResourcesPlugin.getWorkspace()
                        .getRoot();
                fileChanges.fileMoved(workspaceRoot.getFile(
                        delta.getMovedFromPath()).getLocation().toFile(),
                        addedFile);
                return false;
            }
            else
            {
                fileChanges.fileAdded(addedFile);

                // Inform us of files added inside of a directory.
                return true;
            }
        }
        else if (delta.getKind() == IResourceDelta.REMOVED)
        {
            if ((delta.getFlags() & IResourceDelta.MOVED_TO) != 0)
            {
                // We should get an ADDED/MOVED_FROM delta for this too so we
                // don't need to handle a moved to delta.
            }
            else
            {
                IPath removedFileLocation = delta.getResource().getLocation();

                if (removedFileLocation != null)
                {
                    final File removedFile = removedFileLocation.toFile();

                    fileChanges.fileDeleted(removedFile);
                }
            }

            return false;
        }
        else
        {
            return true;
        }
    }

    /**
     * Adds a new set of file changes to the queue of changes to be processed on
     * the background thread.
     */
    private void addFileChanges(FileChanges changes)
    {
        assert changes != null;

        synchronized (mFileChangsQueue)
        {
            mFileChangsQueue.add(changes);
            mFileChangsQueue.notifyAll();
        }
    }

    /**
     * Called on a background thread to process changes to files and execute the
     * appropriate Bazaar commands.
     */
    private void processFileChangsQueue()
    {
        while (!mShouldStop)
        {
            try
            {
                FileChanges changes = null;

                synchronized (mFileChangsQueue)
                {
                    if (mFileChangsQueue.isEmpty())
                    {
                        synchronized (mQueueEmptiedSignal)
                        {
                            mQueueEmptiedSignal.notifyAll();
                        }

                        // Wait for an operation to be added to the queue.

                        mFileChangsQueue.wait();
                    }

                    if (mShouldStop)
                    {
                        break;
                    }

                    changes = mFileChangsQueue.remove(0);
                }

                assert (changes != null);

                processFileChanges(changes);
            }
            catch (InterruptedException e)
            {
                break;
            }
        }
    }

    /**
     * Processes a set of changes to files on the background thread. This will
     * execute the appropriate Bazaar commands to add, delete, and move files.
     */
    private void processFileChanges(FileChanges changes)
    {
        assert changes != null;

        removeUninterestingChanges(changes);

        detectDirectoryRename(changes);

        if (QBzrEclipsePlugin.getInstance().shouldAutoAddFiles()
                && !changes.mAddedFiles.isEmpty())
        {
            // Map that maps from branch directory to the IgnoredFiles object
            // that contains the set of ignored files.
            HashMap<File, IgnoredAndVersionedFiles> ignoredFilesMap = new HashMap<File, IgnoredAndVersionedFiles>();

            for (File addedFile : changes.mAddedFiles)
            {
                addFile(addedFile, ignoredFilesMap);
            }
        }

        for (File deletedFile : changes.mDeletedFiles)
        {
            removeFile(deletedFile);
        }

        for (FileMoveInfo moveInfo : changes.mMovedFiles)
        {
            moveFile(moveInfo.getFrom(), moveInfo.getTo());
        }
    }

    /**
     * Removes changes that are of no interest to us because they are for files
     * that are not inside a branch or should be ignored.
     */
    private void removeUninterestingChanges(FileChanges changes)
    {
        Iterator<File> iterator = changes.mAddedFiles.iterator();

        while (iterator.hasNext())
        {
            File addedFile = iterator.next();

            File branchDirectory = BazaarUtilities
                    .getBranchDirectory(addedFile);

            if (branchDirectory == null)
            {
                iterator.remove();
            }
            else if (branchDirectory.equals(addedFile))
            {
                // A branch has been added. We can ignore any files added under
                // this branch.

                iterator.remove();

                while (iterator.hasNext())
                {
                    addedFile = iterator.next();

                    if (FileUtilities.isParentOf(branchDirectory, addedFile))
                    {
                        iterator.remove();
                    }
                }
            }
        }

        iterator = changes.mDeletedFiles.iterator();

        while (iterator.hasNext())
        {
            final File deletedFile = iterator.next();

            if (BazaarUtilities.getBranchDirectory(deletedFile) == null)
            {
                iterator.remove();
            }
        }
    }

    /**
     * When refactoring in Eclipse and renaming a Java package, Eclipse
     * sometimes sends us this change as a delete of the original directory with
     * an add of the new directory with all files moved inside of the new
     * directory. This method tries to detect this situation by looking for a
     * directory removal and a directory add and all moves being from the source
     * to the destination directory.
     */
    private void detectDirectoryRename(FileChanges changes)
    {
        // For each added directory see if we can find file move operations with
        // the destination being under this directory.

        Iterator<File> iterator = changes.mAddedFiles.iterator();

        while (iterator.hasNext())
        {
            final File addedFile = iterator.next();

            if (addedFile.isDirectory())
            {
                List<FileMoveInfo> filesMovedIntoDirectory = getFilesMovedIntoDirectory(
                        addedFile, changes.mMovedFiles);

                if (!filesMovedIntoDirectory.isEmpty())
                {
                    // Check if we can find a directory delete that is a parent
                    // directory of all these file moves. If we can then assume
                    // this is a rename of this directory.

                    final File deletedDirectory = findDeletedDirectoryForMovedFiles(
                            addedFile, changes.mDeletedFiles,
                            filesMovedIntoDirectory);

                    if (deletedDirectory != null)
                    {
                        // We have detected a directory move:
                        // deletedDirectory -> addedFile

                        changes.mDeletedFiles.remove(deletedDirectory);
                        changes.mMovedFiles.removeAll(filesMovedIntoDirectory);
                        iterator.remove();
                        changes.mMovedFiles.add(new FileMoveInfo(
                                deletedDirectory, addedFile));
                    }
                }
            }
        }
    }

    /**
     * Scans the list of moved files and finds those which were moved into the
     * given directory or one of its sub-directories.
     */
    private List<FileMoveInfo> getFilesMovedIntoDirectory(File directory,
            List<FileMoveInfo> movedFiles)
    {
        List<FileMoveInfo> result = new ArrayList<FileMoveInfo>();

        for (FileMoveInfo fileMoveInfo : movedFiles)
        {
            final File to = fileMoveInfo.getTo();

            if (FileUtilities.isParentOf(directory, to))
            {
                result.add(fileMoveInfo);
            }
        }

        return result;
    }

    /**
     * Tries to find a directory that has been deleted that is a parent/ancestor
     * of all the moved from files.
     * 
     * @return The directory that meets this requirement or null if none.
     */
    private File findDeletedDirectoryForMovedFiles(File toDirectory,
            List<File> deletedFiles,
            List<FileMoveInfo> movedFiles)
    {
        for (File file : deletedFiles)
        {
            if (areAllFileMovesFromOneDirectoryToAnother(file, toDirectory,
                    movedFiles))
            {
                return file;
            }
        }

        return null;
    }

    private boolean areAllFileMovesFromOneDirectoryToAnother(File fromDirectory,
            File toDirectory,
            List<FileMoveInfo> movedFiles)
    {
        assert fromDirectory != null;
        assert toDirectory != null;
        assert movedFiles != null;
        assert !movedFiles.isEmpty();

        for (FileMoveInfo fileMoveInfo : movedFiles)
        {
            final File fromFile = fileMoveInfo.getFrom();
            final File toFile = fileMoveInfo.getTo();

            if (!FileUtilities.isParentOf(fromDirectory, fromFile)
                    || !FileUtilities.isParentOf(toDirectory, toFile))
            {
                return false;
            }

            final String fromPath = FileUtilities.getRelativeUnixPath(
                    fromDirectory, fromFile);
            final String toPath = FileUtilities.getRelativeUnixPath(
                    toDirectory, toFile);

            if (!fromPath.equals(toPath))
            {
                return false;
            }
        }

        return true;
    }

    private void addFile(File file,
            Map<File, IgnoredAndVersionedFiles> ignoredFilesMap)
    {
        assert file != null;

        if (!QBzrEclipsePlugin.getInstance().shouldAutoAdd(file))
        {
            return;
        }

        File branchDirectory = BazaarUtilities.getBranchDirectory(file);

        if (branchDirectory == null)
        {
            return;
        }

        IgnoredAndVersionedFiles ignoredAndVersionedFiles = ignoredFilesMap
                .get(branchDirectory);

        if (ignoredAndVersionedFiles == null)
        {
            ignoredAndVersionedFiles = new IgnoredAndVersionedFiles(
                    branchDirectory);
            ignoredFilesMap.put(branchDirectory, ignoredAndVersionedFiles);
        }

        if (ignoredAndVersionedFiles.hadErrorExecutingCommands()
                || ignoredAndVersionedFiles.isIgnored(file)
                || ignoredAndVersionedFiles.isVersioned(file))
        {
            return;
        }

        String[] command = new String[]
        { QBzrEclipsePlugin.getInstance().getBzrExecutablePath(), "add",
                "--no-aliases",
                FileUtilities.getRelativePath(branchDirectory, file) };

        QBzrEclipsePlugin.getInstance().logCommand(command);

        try
        {
            ProcessUtilities.exec(command, branchDirectory);
        }
        catch (ProcessException e)
        {
            QBzrEclipsePlugin.getInstance().logException("Could not add file",
                    e);
        }
    }

    private void removeFile(File file)
    {
        assert file != null;

        // Don't remove files from Bazaar.
        // The main reason is that if the user deletes a file in Eclipse and
        // then
        // performs undo we don't want the file to appear as removed and added
        // in Bazaar.
        // Bazaar also does not require us to use bzr remove command and will
        // assume
        // any file not present during commit has been removed.
    }

    /** Performs a Bazaar move command to move the file to its new location. */
    private void moveFile(File from, File to)
    {
        assert from != null;
        assert to != null;

        final File branchDirectory = BazaarUtilities.getBranchDirectory(from);
        final File toBranchDirectory = BazaarUtilities.getBranchDirectory(to);

        if (branchDirectory == null && toBranchDirectory == null)
        {
            // Neither to or from files are under Bazaar version control so
            // ignore this move.

            return;
        }

        if (branchDirectory == null
                || !branchDirectory.equals(toBranchDirectory))
        {
            // This move was not within one branch. We should remove the file
            // from the source branch and add it in the target branch.

            if (branchDirectory != null)
            {
                removeFile(from);
            }

            if (toBranchDirectory != null)
            {
                // addFile(to);
            }

            return;
        }

        if (!BazaarUtilities.isVersioned(from))
        {
            return;
        }

        // Make sure parent directories are added if we moved into a
        // different directory.

        if (!BazaarUtilities.addDirectoriesNeededForFile(branchDirectory, to))
        {
            return;
        }

        String[] command = new String[]
        { QBzrEclipsePlugin.getInstance().getBzrExecutablePath(), "move",
                "--no-aliases", "--after",
                FileUtilities.getRelativePath(branchDirectory, from),
                FileUtilities.getRelativePath(branchDirectory, to) };

        QBzrEclipsePlugin.getInstance().logCommand(command);

        try
        {
            ProcessUtilities.exec(command, branchDirectory);
        }
        catch (ProcessException e)
        {
            QBzrEclipsePlugin.getInstance().logException("Could not move file",
                    e);
        }
    }

    /** Holds information about where a file was moved to. */
    private static final class FileMoveInfo
    {
        public FileMoveInfo(File from, File to)
        {
            assert from != null;
            assert to != null;

            this.mFrom = from;
            this.mTo = to;
        }

        public File getFrom()
        {
            return mFrom;
        }

        public File getTo()
        {
            return mTo;
        }

        private final File mFrom, mTo;
    }

    /**
     * Defines a set of changes that need to be processed by the background
     * thread.
     */
    private static final class FileChanges
    {
        public void fileAdded(File file)
        {
            mAddedFiles.add(file);
        }

        public void fileDeleted(File file)
        {
            mDeletedFiles.add(file);
        }

        public void fileMoved(File from, File to)
        {
            mMovedFiles.add(new FileMoveInfo(from, to));
        }

        public final List<File> mAddedFiles = new ArrayList<File>();

        public final List<File> mDeletedFiles = new ArrayList<File>();

        public final List<FileMoveInfo> mMovedFiles = new ArrayList<FileMoveInfo>();
    }

    private final Thread mThread;

    private boolean mShouldStop;

    /** Queue that is processed by the background thread. */
    private final List<FileChanges> mFileChangsQueue = new LinkedList<FileChanges>();

    private final Object mQueueEmptiedSignal = new Object();
}
