package org.bazaar_vcs.qbzr_eclipse.handlers;

import java.io.File;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.bazaar_vcs.qbzr_eclipse.ResourceChangeListener;
import org.bazaar_vcs.qbzr_eclipse.QBzrEclipsePlugin;
import org.bazaar_vcs.qbzr_eclipse.preferences.PreferenceConstants;
import org.bazaar_vcs.qbzr_eclipse.util.BazaarUtilities;
import org.bazaar_vcs.qbzr_eclipse.util.EclipseUtilities;
import org.bazaar_vcs.qbzr_eclipse.util.FileUtilities;
import org.bazaar_vcs.qbzr_eclipse.util.ProcessException;
import org.bazaar_vcs.qbzr_eclipse.util.ProcessResult;
import org.bazaar_vcs.qbzr_eclipse.util.ProcessUtilities;
import org.bazaar_vcs.qbzr_eclipse.util.StringUtilities;
import org.eclipse.core.commands.AbstractHandler;
import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.ITreeSelection;
import org.eclipse.swt.widgets.Display;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.handlers.HandlerUtil;

/**
 * Base class for command handlers that invoke a Bazaar command.
 */
public abstract class BazaarCommandHandler extends AbstractHandler
{
	protected BazaarCommandHandler()
	{
	}

	/**
	 * the command has been executed, so extract extract the needed information
	 * from the application context.
	 */
	@SuppressWarnings("unchecked")
	public Object execute(final ExecutionEvent event) throws ExecutionException
	{

		if (shouldSaveBeforeExecuting())
		{
			IWorkbenchWindow workbenchWindow = HandlerUtil
					.getActiveWorkbenchWindowChecked(event);

			if (!workbenchWindow.getActivePage().saveAllEditors(true))
			{
				return null;
			}
		}

		// Work out what files are selected and which branch directory we are
		// in.

		ISelection selection = HandlerUtil.getCurrentSelection(event);

		List<File> selectedFiles = Collections.EMPTY_LIST;
		File branchDirectory = null;
		File fileBeingEdited = null;

		if (selection != null && !selection.isEmpty() && selection instanceof ITreeSelection)
		{
			selectedFiles = EclipseUtilities
					.getSelectedFiles((ITreeSelection) selection);
		}		
		
		if (selectedFiles.isEmpty())
		{
			fileBeingEdited = EclipseUtilities.getFileBeingEdited(event);

			branchDirectory = BazaarUtilities
					.getBranchDirectory(fileBeingEdited);
		}
		else
		{
		    branchDirectory = BazaarUtilities.getBranchDirectory(selectedFiles);
		}

		// Check that the command can run on the selection.

		try
		{
			checkCommandCanRunOnSelection(selectedFiles, fileBeingEdited,
					branchDirectory);
		}
		catch (CannotRunCommandOnSelectionException e)
		{
			showError(e.getLocalizedMessage(), event);
			return null;
		}

		// Build the command and execute it.

		IPath workspaceRoot = ResourcesPlugin.getWorkspace().getRoot()
				.getLocation();

		final File directoryToRunIn = getDirectoryToRunIn(workspaceRoot
				.toFile(), branchDirectory, selectedFiles, fileBeingEdited);

		List<String> commandElements = getCommand(selectedFiles,
				fileBeingEdited, branchDirectory, directoryToRunIn);
		final String[] command = commandElements
				.toArray(new String[commandElements.size()]);

		QBzrEclipsePlugin.getInstance().logCommand(command);

		// Execute the command in a background thread so as to not lock up the
		// user interface thread.

		new Thread()
		{
			@Override
			public void run()
			{
				try
				{
					if (shouldWaitForFilesToBeMovedAndDeleted())
					{
						ResourceChangeListener.INSTANCE
								.waitForFilesToBeMovedAndDeleted();
					}

					final ProcessResult result = ProcessUtilities.exec(command,
							directoryToRunIn);
					
					// this must get called on failure or success. 
					// Reason for on failure is that update may apparently fail because of conflicts
					// yet it changed files
					if(result.getExitCode() == 0 &&
							shouldRefreshWorkspace() && QBzrEclipsePlugin.getInstance().getPreferenceStore().getBoolean(
							PreferenceConstants.REFRESH_PROJECTS_AFTER_FILE_CHANGING_COMMANDS))
					{
						refreshFiles(directoryToRunIn);
					}
					
					Display.getDefault().asyncExec(new Runnable()
					{
						public void run()
						{
							commandCompleted(result, event);
							
							if (mListener != null)
							{
								if (isSuccessfulExitCode(result.getExitCode()))
								{
									mListener.commandSucceeded(result, event);
								}
								else if (isAbortExitCode(result.getExitCode()))
								{
								    mListener.commandAborted(result, event);
								}
								else
								{
									mListener.commandFailed(result, event);
								}
							}
						}
					});
				}
				catch (final ProcessException e)
				{
					Display.getDefault().asyncExec(new Runnable()
					{
						public void run()
						{
							commandCouldNotBeExecuted(e, event);

							if (mListener != null)
							{
								mListener.commandCouldNotBeExecuted(e, event);
							}
						}
					});
				}
			}

		}.start();

		return null;
	}

	/**
	 * Gets the directory this command should be run in. By default, if no
	 * branch is found then the command will run in the workspace root
	 * directory.
	 * 
	 * @param workspaceRootDir
	 *            The directory that represents the root of the workspace.
	 * @param branchDir
	 *            The detected branch directory of the selected files or null if
	 *            none.
	 * @return The directory to run in.
	 */
	protected File getDirectoryToRunIn(File workspaceRootDir,
			File branchDir,
			List<File> selectedFiles,
			File fileBeingEdited)
	{
		assert workspaceRootDir != null;

		return branchDir == null ? workspaceRootDir : branchDir;
	}

	protected boolean shouldSaveBeforeExecuting()
	{
		return false;
	}

	protected boolean shouldWaitForFilesToBeMovedAndDeleted()
	{
		return true;
	}

	/** Checks if this command requires a branch in order to execute. */
	protected boolean requiresBranchDirectory()
	{
		return true;
	}

	/**
	 * Checks if the list of selected files should be passed on to the Bazaar
	 * command.
	 */
	protected boolean useSelectedFilesAsArguments()
	{
		return true;
	}

	/**
	 * Sub-classes can override this to validate the selection prior to
	 * executing the command. If the selection is invalid then should throw
	 * CannotRunCommandOnSelectionException with a message explaining why not.
	 * 
	 * @param selectedFiles
	 *            The files selected in the browser
	 * @param fileBeingEdited
	 *            The file being edited in the active editor.
	 * @param branchDirectory
	 *            The Bazaar branch directory that contains the selection or
	 *            null if none.
	 * @throws CannotRunCommandOnSelectionException
	 *             if selection is not valid.
	 */
	protected void checkCommandCanRunOnSelection(List<File> selectedFiles,
			File fileBeingEdited,
			File branchDirectory) throws CannotRunCommandOnSelectionException
	{
		if (requiresBranchDirectory() && branchDirectory == null)
		{
			throw new CannotRunCommandOnSelectionException(
					"The current editor or selection does not contain any items under Bazaar version control");
		}
	}

	/**
	 * Gets the command that should be run. The first item should be the bzr
	 * executable and the following items the arguments to the command.
	 * 
	 * @param selectedFiles
	 *            The files currently selected (e.g. in the browser).
	 * @param fileBeingEdited
	 *            The file that is currently being edited in the text editor (or
	 *            null if none).
	 * @param branchDirectory
	 *            The directory that contains the Bazaar branch.
	 */
	protected List<String> getCommand(List<File> selectedFiles,
			File fileBeingEdited,
			File branchDirectory,
			File directoryToRunIn)
	{
		List<String> command = new ArrayList<String>();

		command.add(QBzrEclipsePlugin.getInstance().getBzrExecutablePath());
		command.add(getCommandName());
		
		if(isUIModeSupported() && QBzrEclipsePlugin.getInstance().shouldLeaveDialogOpen())
		{
			command.add("--ui-mode");
		}
		
		command.add("--no-aliases");

		addArguments(command, selectedFiles, fileBeingEdited, branchDirectory,
				directoryToRunIn);

		return command;
	}
	
	protected boolean isUIModeSupported(){
		return false;
	}

	/**
	 * Adds any arguments needed by this command. The default will add the
	 * selected files as arguments if that makes sense for this command (@see
	 * useSelectedFilesAsArguments).
	 */
	protected void addArguments(List<String> command,
			List<File> selectedFiles,
			File fileBeingEdited,
			File branchDirectory,
			File directoryToRunIn)
	{
		if (useSelectedFilesAsArguments())
		{
			for (File file : selectedFiles)
			{
				command.add(FileUtilities.getRelativePath(directoryToRunIn,
						file));
			}
		}
	}

	/** The name of the command that is passed to Bazaar. */
	protected abstract String getCommandName();

	/**
	 * Checks if the exit code of the command indicates success or not. By
	 * default a successful exit code is 0.
	 */
	protected boolean isSuccessfulExitCode(int exitCode)
	{
		return exitCode == 0;
	}
	
	/**
     * Checks if the exit code of the command indicates if the command was aborted (e.g Cancel button pressed).
     * By default an exit code of 1 indicates the command was aborted.
     */
    protected boolean isAbortExitCode(int exitCode)
    {
        return exitCode == 1;
    }

	/**
	 * This is called if it is not possible to execute the command. This is
	 * always called on the user interface thread.
	 */
	protected void commandCouldNotBeExecuted(ProcessException e,
			ExecutionEvent event)
	{
		showError("Error executing command\n" + e.getMessage(), event);
	}

	/**
	 * This is called when the bzr command completes. This is always called on
	 * the user interface thread.
	 */
	protected final void commandCompleted(ProcessResult result, ExecutionEvent event)
	{
		if (isSuccessfulExitCode(result.getExitCode()))
		{
			commandSucceeded(result, event);
		}
		else if (isAbortExitCode(result.getExitCode()))
		{
		    commandAborted(result, event);
		}
		else
		{
			commandFailed(result, event);
		}
	}

	protected boolean shouldRefreshWorkspace()
	{
		return false;
	}
	
    private void refreshFiles(final File directoryToRunIn)
    {
        final Job refresh = new Job("Refreshing Projects")
        {
            @Override
            protected IStatus run(IProgressMonitor monitor)
            {
                final IProject[] prjs = ResourcesPlugin.getWorkspace()
                        .getRoot().getProjects();
                monitor.beginTask("Refreshing Projects", prjs.length);
                final String path = directoryToRunIn.toString();
                
                for (IProject project : prjs)
                {
                    if (project.isOpen()
                            && project.getLocation().equals(new Path(path))
                            || project.getLocation().toOSString()
                                    .contains(path))
                    {
                        try
                        {
                            project.refreshLocal(IResource.DEPTH_INFINITE,
                                    monitor);
                        }
                        catch (CoreException e)
                        {
                            QBzrEclipsePlugin
                                    .getInstance()
                                    .getLog()
                                    .log(
                                            new Status(
                                                    IStatus.ERROR,
                                                    "org.bazaar_vcs.qbzr_eclipse.handlers",
                                                    "Failed to refresh project: "
                                                            + project.getName(),
                                                    e));
                        }
                    }
                    monitor.worked(1);
                }
                monitor.done();
                return Status.OK_STATUS;
            }
        };
        refresh.schedule();
    }
	
	protected void commandSucceeded(ProcessResult result, ExecutionEvent event)
	{
		
	}
	
	protected void commandAborted(ProcessResult result, ExecutionEvent event)
    {

    }

	protected void commandFailed(ProcessResult result, ExecutionEvent event)
	{
		String message = "Error executing command: "
				+ StringUtilities.createStringWithQuotes(result.getCommand())
				+ "\n" + result.getErrorMessages();
		QBzrEclipsePlugin.getInstance().logError(message);
		showError(message, event);
	}

	/** Shows an information message to the user. */
	protected void showInformation(String message, ExecutionEvent event)
	{
		try
		{
			IWorkbenchWindow window = HandlerUtil
					.getActiveWorkbenchWindowChecked(event);
			MessageDialog.openInformation(window.getShell(),
					"QBzr Eclipse Plug-in", message);
		}
		catch (ExecutionException e)
		{
			e.printStackTrace();
		}
	}

	/** Shows an error message to the user. */
	protected void showError(String message, ExecutionEvent event)
	{
		try
		{
			IWorkbenchWindow window = HandlerUtil
					.getActiveWorkbenchWindowChecked(event);
			MessageDialog.openError(window.getShell(), "QBzr Eclipse Plug-in",
					message);
		}
		catch (ExecutionException e)
		{
			e.printStackTrace();
		}
	}

	public void setListener(BazaarCommandHandlerListener listener)
	{
		assert (mListener == null);

		mListener = listener;
	}

	private BazaarCommandHandlerListener mListener;
}
