package org.bazaar_vcs.qbzr_eclipse.util;

import java.io.File;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.bazaar_vcs.qbzr_eclipse.QBzrEclipsePlugin;

/** A collection of useful utilities to perform Bazaar operations and queries. */

public final class BazaarUtilities
{
    /**
     * Gets the branch directory for a given file or null if this file is not
     * inside a branch.
     */
    public static File getBranchDirectory(File file)
    {
        while (file != null)
        {
            if (new File(file, ".bzr" + File.separator + "branch").isDirectory())
            {
                return file;
            }

            file = file.getParentFile();
        }

        return null;
    }

    /**
     * Gets the branch directory for a collection of files or null if not all
     * files are inside the same branch.
     */
    public static File getBranchDirectory(Collection<File> files)
    {
        File branchDir = null;

        for (File file : files)
        {
            File currentBranchDir = getBranchDirectory(file);

            if (currentBranchDir == null)
            {
                return null;
            }

            if (branchDir == null)
            {
                branchDir = currentBranchDir;
            }
            else if (!branchDir.equals(currentBranchDir))
            {
                return null;
            }
        }

        return branchDir;
    }

    /**
     * Checks if the supplied file is under Bazaar version control (i.e exists
     * in the repository).
     */
    public static boolean isVersioned(File file)
    {
        String[] command = new String[]
        { QBzrEclipsePlugin.getInstance().getBzrExecutablePath(),
                "--no-aliases", "is-versioned", file.getAbsolutePath() };

        try
        {
            QBzrEclipsePlugin.getInstance().logCommand(command);
            ProcessResult result = ProcessUtilities.exec(command, null);

            return result.getExitCode() == 1;
        }
        catch (ProcessException e)
        {
            QBzrEclipsePlugin.getInstance().logException(
                    "Could not check if file is under Bazaar version control",
                    e);
        }

        return false;
    }

    /**
     * Adds any parent directories that are not already added to Bazaar so that
     * the given file can be added.
     */
    public static boolean addDirectoriesNeededForFile(File branchDirectory,
            File file)
    {
        List<File> dirsToAdd = new ArrayList<File>();

        File parentDir = file.getParentFile();

        while (!parentDir.equals(branchDirectory)
                && !BazaarUtilities.isVersioned(parentDir))
        {
            dirsToAdd.add(0, parentDir);
            parentDir = parentDir.getParentFile();
        }

        for (File dirToAdd : dirsToAdd)
        {
            String[] command = new String[]
            { QBzrEclipsePlugin.getInstance().getBzrExecutablePath(), "add",
                    "--no-aliases", "--no-recurse",
                    FileUtilities.getRelativePath(branchDirectory, dirToAdd) };

            QBzrEclipsePlugin.getInstance().logCommand(command);

            try
            {
                ProcessResult result = ProcessUtilities.exec(command,
                        branchDirectory);

                if (result.getExitCode() != 0)
                {
                    return false;
                }
            }
            catch (ProcessException e)
            {
                QBzrEclipsePlugin.getInstance().logException(
                        "Could not add directory", e);
                return false;
            }
        }

        return true;
    }

    private static class IgnoreStatus
    {
        public IgnoreStatus(long lastCheckTime, boolean isIgnored)
        {
            super();
            this.mLastCheckTime = lastCheckTime;
            this.mIsIgnored = isIgnored;
        }

        public long mLastCheckTime;

        public boolean mIsIgnored;
    }

    /**
     * Maximum time that a file known to be ignored should remain in the ignored
     * file cache.
     */
    private static final int MAX_IGNORE_CACHE_TIME = 10000;

    /**
     * Cache of files that are known to be ignored. The value in the cache is
     * the last time that the file was known to be ignored.
     */
    private static final Map<File, IgnoreStatus> gIgnoredFilesCache = new HashMap<File, IgnoreStatus>();
}
