package org.bazaar_vcs.qbzr_eclipse;


import static org.junit.Assert.assertTrue;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.StringTokenizer;

import org.bazaar_vcs.qbzr_eclipse.util.BazaarUtilities;
import org.bazaar_vcs.qbzr_eclipse.util.ProcessException;
import org.bazaar_vcs.qbzr_eclipse.util.ProcessResult;
import org.bazaar_vcs.qbzr_eclipse.util.ProcessUtilities;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IFolder;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IProjectDescription;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IWorkspaceRoot;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;

public class ResourceChanges {
	private IWorkspaceRoot root;
	private IProject project;
	private IFolder bzrFolder;

	@Before
	public void setup(){
		try{
			root = ResourcesPlugin.getWorkspace().getRoot();
			project = root.getProject("test");
			project.create(null);
			project.open(null);
			final IProjectDescription description = project.getDescription();
			description.setNatureIds(new String[0]);
			project.setDescription(description, null);
			bzrFolder = project.getFolder("Branch");
			if (!bzrFolder.exists()){
				bzrFolder.create(IResource.NONE, true, null);
			}
			init(bzrFolder);
			final File branchDirectory = BazaarUtilities.getBranchDirectory(bzrFolder.getLocation().toFile());
			assertTrue("Expected a initialized bazaar dir: "+
					bzrFolder.getLocation().toFile().toString(), branchDirectory!=null);
		} catch (CoreException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	@After
	public void tearDown(){
		try{
			bzrFolder.delete(true, null);
			project.delete(true, null);
		} catch (CoreException e) {
			e.printStackTrace();
		}
	}

	@Test
	public void testRenamingSingleFile(){
		final IFile file = bzrFolder.getFile("file1.txt");
		try{
			if (!file.exists()) {
				final byte[] bytes = "File contents".getBytes();
				final InputStream source = new ByteArrayInputStream(bytes);
				file.create(source, IResource.FORCE, null);
			}
			addFiles();
			commitAllChanges();
			file.refreshLocal(IFile.DEPTH_INFINITE, null);
			final IFile file2 = bzrFolder.getFile("file2.txt");
			file.move(file2.getFullPath(), true, null);
			file.refreshLocal(IFile.DEPTH_INFINITE, null);
			ResourceChangeListener.INSTANCE.waitForFilesToBeMovedAndDeleted();
			
			List<String> out = getStatus();
			assertTrue("Rename was expected", out.size() > 0 && out.get(0).equals("renamed:"));
			final String result = "file1.txt => file2.txt";
			assertTrue("Expected another value but only one found",out.size() > 1);
			assertTrue("The following change was expected: "+result+ 
					" but the following was recieved: " + out.get(1)
					,out.get(1).equals(result));
		} catch (CoreException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	@Test
	public void testMoveSingleChildDirectory(){
		try{
			IFolder folderFrom = bzrFolder.getFolder("FolderFROM");
			if (!folderFrom.exists()){
				folderFrom.create(IResource.NONE, true, null);
			}
			addFiles();
			commitAllChanges();
			bzrFolder.refreshLocal(IFile.DEPTH_INFINITE, null);

			IFolder folderTo = bzrFolder.getFolder("FolderTO");
			folderFrom.move(folderTo.getFullPath(), true, null);
			bzrFolder.refreshLocal(IFile.DEPTH_INFINITE, null);
			ResourceChangeListener.INSTANCE.waitForFilesToBeMovedAndDeleted();
			
			final List<String> out = getStatus();
			assertTrue("Rename was expected", out.size() > 0 && out.get(0).equals("renamed:"));
			final String result = "FolderFROM/ => FolderTO/";
			assertTrue("Expected another value but only one found",out.size() > 1);
			assertTrue("The following change was expected: "+result+ 
					" but the following was recieved: " + out.get(1)
					,out.get(1).equals(result));
		} catch (CoreException e) {
			e.printStackTrace();
		}
	}
	
	@Test
	public void testMoveSingleParentDirectory(){
		try{
			IFolder folderFromParent  = bzrFolder.getFolder("FolderFromParent");
			if (!folderFromParent.exists()){
				folderFromParent.create(IResource.NONE, true, null);
			}
			IFolder folderFromChild = folderFromParent.getFolder("FolderFromChild");
			if (!folderFromChild.exists()){
				folderFromChild.create(IResource.NONE, true, null);
			}
			
			addFiles();
			commitAllChanges();
			bzrFolder.refreshLocal(IFile.DEPTH_INFINITE, null);

			IFolder folderTo = bzrFolder.getFolder("FolderTO");
			folderFromParent.move(folderTo.getFullPath(), true, null);
			bzrFolder.refreshLocal(IFile.DEPTH_INFINITE, null);
			ResourceChangeListener.INSTANCE.waitForFilesToBeMovedAndDeleted();
			
			final List<String> out = getStatus();
			
			assertTrue("Rename was expected", out.size() > 0 && out.get(0).equals("renamed:"));
			final String result = "FolderFromParent/ => FolderTO/";
			assertTrue("Expected another value but only one found",out.size() > 1);
			assertTrue("The following change was expected: "+result+ 
					" but the following was recieved: " + out.get(1)
					,out.get(1).equals(result));
		} catch (CoreException e) {
			e.printStackTrace();
		}
	}
	
	@Test
	public void testMoveSingleParentDirectory_WithChildren(){
		try{
			IFolder folderFromParent  = bzrFolder.getFolder("FolderFromParent");
			if (!folderFromParent.exists()){
				folderFromParent.create(IResource.NONE, true, null);
			}
			IFolder folderFromChild = folderFromParent.getFolder("FolderFromChild");
			if (!folderFromChild.exists()){
				folderFromChild.create(IResource.NONE, true, null);
			}
			
			final IFile file = folderFromChild.getFile("file1.txt");
			if (!file.exists()) {
				final byte[] bytes = "File contents".getBytes();
				final InputStream source = new ByteArrayInputStream(bytes);
				file.create(source, IResource.FORCE, null);
			}
			final IFile file2 = folderFromChild.getFile("file2.txt");
			if (!file2.exists()) {
				final byte[] bytes = "File contents".getBytes();
				final InputStream source = new ByteArrayInputStream(bytes);
				file2.create(source, IResource.FORCE, null);
			}
				
			addFiles();
			commitAllChanges();
			bzrFolder.refreshLocal(IFile.DEPTH_INFINITE, null);

			IFolder folderTo = bzrFolder.getFolder("FolderTO");
			folderFromParent.move(folderTo.getFullPath(), true, null);
			bzrFolder.refreshLocal(IFile.DEPTH_INFINITE, null);
			ResourceChangeListener.INSTANCE.waitForFilesToBeMovedAndDeleted();
			
			final List<String> out = getStatus();
			assertTrue("Rename was expected", out.size() > 0 && out.get(0).equals("renamed:"));
			final String result = "FolderFromParent/ => FolderTO/";
			assertTrue("Expected another value but only one found",out.size() > 1);
			assertTrue("The following change was expected: "+result+ 
					" but the following was recieved: " + out.get(1)
					,out.get(1).equals(result));
		} catch (CoreException e) {
			e.printStackTrace();
		}
	}
	
	//~~~~~~~~~~~~~~ Utility methods ~~~~~~~~~~~~~~~~~~`
	
//	private void printList(List<String> commands){
//		for(String o: commands){
//			System.out.println(o);
//		}
//	}
	
	
	private void init(IFolder folder){
		String[] command = new String[]{ QBzrEclipsePlugin.getInstance().getBzrExecutablePath(), "init"};
		QBzrEclipsePlugin.getInstance().logCommand(command);
		try{
			ProcessUtilities.exec(command, folder.getLocation().toFile());
		}catch (ProcessException e){
			QBzrEclipsePlugin.getInstance().logException(
					"Could not add file", e);
		}
	}
	private void commitAllChanges(){
		String[] command = new String[]{ QBzrEclipsePlugin.getInstance().getBzrExecutablePath(), "commit",
				"-m","test"};
		QBzrEclipsePlugin.getInstance().logCommand(command);
		try{
			ProcessUtilities.exec(command, bzrFolder.getLocation().toFile());
		}catch (ProcessException e){
			QBzrEclipsePlugin.getInstance().logException(
					"Could not add file", e);
		}
	}
	private void addFiles(){
		String[] command = new String[]{ QBzrEclipsePlugin.getInstance().getBzrExecutablePath(), "add"};
		QBzrEclipsePlugin.getInstance().logCommand(command);
		try{
			ProcessUtilities.exec(command, bzrFolder.getLocation().toFile());
		}catch (ProcessException e){
			QBzrEclipsePlugin.getInstance().logException(
					"Could not add file", e);
		}
	}

	private List<String> getStatus(){
		final String[] command = new String[]{ 
				QBzrEclipsePlugin.getInstance().getBzrExecutablePath(), "status"};
		QBzrEclipsePlugin.getInstance().logCommand(command);
		try{
			final ProcessResult result = ProcessUtilities.exec(command, bzrFolder.getLocation().toFile());
			final List<String> out = new ArrayList<String>();
			final StringTokenizer tk = new StringTokenizer(result.getOutputMessages(), "\n");
			while(tk.hasMoreElements()){
				out.add(((String)(tk.nextElement())).trim());
			}
			return out;
		}catch (ProcessException e){
			QBzrEclipsePlugin.getInstance().logException(
					"Could not get status", e);
		}	
		return null;
	}
}